package de.hawhamburg.reachability.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathException;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Utility functions related to XML handling.<br><br>
 * 
 * This utility class provides some fancy utility methods related to XML handling.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class XMLUtil {

    /**
     * Prevent instances of this utility class.
     */
    private XMLUtil() { }
    
    /**
     * Loads an XML document from the file the supplied path refers to. 
     * 
     * @param path      The path referred to the file which existence shall be read as XML document.
     * @return          The {@link Document} representing the XML files content, otherwise <code>null</code>.
     */
    public static Document loadDocument(final String path) {
        Document result = null;
        
        if (ExternFileUtil.exists(path)) {
            File xmlFile = new File(path);
            try {
                DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                result                  = builder.parse(xmlFile);
                
                result.getDocumentElement().normalize();
            } catch (FactoryConfigurationError e) {
                OutputUtil.error("[FactoryConfigurationError] "     + e.getMessage());
            } catch (ParserConfigurationException e) {
                OutputUtil.error("[ParserConfigurationException] "  + e.getMessage());
            } catch (IOException e) {
                OutputUtil.error("[IOException] "                   + e.getMessage());
            } catch (SAXException e) {
                OutputUtil.error("[SAXException] "                  + e.getMessage());
            }
        } else {
            OutputUtil.error("Requested path is not existing '" + path + "'");
        }
        
        return result;
    }
    
    /**
     * Fetches a {@link List} of {@link Node} from the supplied {@link Document} using the given xPath.
     * 
     * @param doc       The XML document as {@link Document}.
     * @param xpath     The xPath expression as {@link String}.
     * @return          The result of the request as {@link List} of {@link Node}.
     */
    public static List<Node> nodeListByXPath(final Document doc,
                                             final String xpath) {
        String     localXpath = xpath;
        List<Node> result     = new ArrayList<Node>();
        
        if (doc != null) {
            if (localXpath == null) {
                localXpath = "";
            }
            
            XPath xp = XPathFactory.newInstance().newXPath();
            try {
                XPathExpression xpr = xp.compile(localXpath);
                NodeList temp = (NodeList) xpr.evaluate(doc, XPathConstants.NODESET);
                for (int i = 0; i < temp.getLength(); i++) {
                    result.add(temp.item(i));
                }
            } catch (XPathException e) {
                OutputUtil.error("[XPathException] " + e.getMessage() + " (Parms: doc='" + doc
                        + "', xpath='" + xpath + "')");
            } catch (NullPointerException e) {
                OutputUtil.error("[NullPointerException] " + e.getMessage() + " (Parms: doc='" + doc
                        + "', xpath='" + xpath + "')");
            }
        } else {
            OutputUtil.error("The supplied document was null for xpath '" + xpath + "'!");
        }
        
        return result;
    }
    

    /**
     * Fetches a {@link List} of {@link Node} from the supplied {@link Node} using the given xPath.
     * 
     * @param node      The XML node as {@link Node}.
     * @param xpath     The xPath expression as {@link String}.
     * @return          The result of the request as {@link List} of {@link Node}.
     */
    public static List<Node> nodeListByXPath(final Node node,
                                             final String xpath) {
        String     localXpath = xpath;
        List<Node> result     = new ArrayList<Node>();

        if (node != null && node.hasChildNodes()) {
            if (localXpath == null) {
                localXpath = "";
            }
            
            XPath xp = XPathFactory.newInstance().newXPath();
            try {
                XPathExpression xpr = xp.compile(localXpath);
                NodeList temp = (NodeList) xpr.evaluate(node, XPathConstants.NODESET);
                for (int i = 0; i < temp.getLength(); i++) {
                    result.add(temp.item(i));
                }
            } catch (XPathException e) {
                System.err.println("[XPathException] " + e.getMessage() + " (Parms: node='" + node
                                   + "', xpath='" + xpath + "')");
            } catch (NullPointerException e) {
                OutputUtil.error("[NullPointerException] " + e.getMessage() + " (Parms: node='"
                                 + node + "', xpath='" + xpath + "')");
            }
        } else {
            OutputUtil.error("The supplied node was null for xpath '" + xpath + "'!");
        }
        
        return result;
    }
    

    /**
     * Fetches a {@link List} of {@link Element} from the supplied {@link Node} using the given xPath.
     * 
     * @param node      The XML node as {@link Node}.
     * @param xpath     The xPath expression as {@link String}.
     * @return          The result of the request as {@link List} of {@link Element}.
     */
    public static List<Element> elementListByXPath(final Node node,
                                                   final String xpath) {
        List<Element> result = new ArrayList<Element>();
        for (Node x : XMLUtil.nodeListByXPath(node, xpath)) {
            result.add((Element) x);
        }
        return result;
    }

    /**
     * Fetches text value as {@link String} from the supplied {@link Document} using the given xPath.
     * 
     * @param doc       The XML document as {@link Document}.
     * @param xpath     The xPath expression as {@link String}.
     * @return          The text value as result of the request as {@link String}.
     */
    public static String getTextContent(final Document doc,
                                        final String xpath) {
        String result = null;
        List<Element> tmp = XMLUtil.elementListByXPath(doc, xpath);
        if (tmp.size() > 0) {
            result = tmp.get(0).getTextContent();
        }
        return result;
    }

    /**
     * Fetches value as integer from the supplied {@link Document} using the given xPath.
     * 
     * @param doc       The XML document as {@link Document}.
     * @param xpath     The xPath expression as {@link String}.
     * @return          The value as result of the request as integer.
     */
    public static int getIntegerContent(final Document doc,
                                        final String xpath) {
        String result = null;
        List<Element> tmp = XMLUtil.elementListByXPath(doc, xpath);
        if (tmp.size() > 0) {
            result = tmp.get(0).getTextContent();
        }
        return Integer.valueOf(result);
    }
    
    /**
     * Fetches the root node from the given {@link Document} and returns it.
     * 
     * @param doc       The XML document as {@link Document}.
     * @return          The root node as {@link Node}.
     */
    public static Node getRoot(final Document doc) {
        return doc.getDocumentElement();
    }
}
